#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Copyright 2024 The SVUT Authors

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
of the Software, and to permit persons to whom the Software is furnished to do
so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

# pylint: disable=C0103

import os
import sys
import re
from string import Template
from pathlib import Path

SCRIPTDIR = os.path.abspath(os.path.dirname(os.path.realpath(__file__)))

def parse_verilog(verilog):

    """
    The next section will parse the user file and extract the parameters
    list, the IOs name and width and the module name. Expect Verilog 2005
    style ONLY for IOs!
    This FSM is a very basic parser and handle simple construction.
    TODO: replace it with serious implementation like pyverilog
    """

    print("INFO: Extract information from module to test")

    # List of flags activated during parsing steps.
    intoComment = "No"
    inlineComment = "Yes"
    moduleFound = "No"
    parameterFound = "No"
    ioFound = "No"

    instance = {"name": "", "io": [], "parameter": []}

    # One by one the loop will detect the different part of the
    # module declaration. Steps are:
    # 1. Detect block comments in file header
    # 2. Detect module name
    # 3. Detect parameters
    # 4. Detect input/output of the module
    # Once detected, store the result for testsuite writing

    for line in verilog:

        # Remove space at beginning and end of line
        line = line.strip()

        # Detect comment block in header and avoid to parse them
        # A block comment like a license is thus ignored
        if line[0:2] == "/*":
            intoComment = "Yes"
            inlineComment = "No"
        elif line[0:2] == "//":
            intoComment = "Yes"
            inlineComment = "Yes"
        elif line[0:2] == "*/" or line[-2:] == "*/":
            intoComment = "No"
            inlineComment = "No"

        if intoComment == "Yes":
            if inlineComment == "Yes":
                intoComment = "No"
            continue

        # Search for the module name if `module` found, split line with " " and
        # get the last part, the name.  Expect `module module_name` on the line
        if moduleFound == "No":
            if "module" in line:
                moduleFound = "Yes"
                info = line.split(" ")
                instance["name"] = info[1]
                if instance["name"][-1] == ";":
                    instance["name"] = instance["name"][:-1]

        # Search for the parameter if present search a line with `parameter`,
        # remove comment at the end of line, replace comma with semicolon and
        # store the line, ready to be written as a parameter declaration in
        # testsuite file
        if parameterFound == "No":
            if line[0:9] == "parameter":
                _line = line.split("//")[0].strip()
                _line = _line.replace("\t", " ")
                _line = _line.replace(",", "")
                if _line[-1] != ";":
                    _line = _line + ";"
                instance["parameter"].append(_line)

        # Search for input or ouput, change comma to semicolon, signed|wire to
        # reg and remove IO mode. Remove comment at the end of line
        # Ready to be written into testsuitefile.
        if ioFound == "No":
            if line[0:5] == "input" or line[0:6] == "output":
                _line = line.split("//")[0].strip()
                if line[0:10] == "input var ":
                    _line = re.sub("input var", "", _line)
                else:
                    _line = re.sub("input", "", _line)
                _line = re.sub("output", "", _line)
                _line = re.sub("signed", "logic", _line)
                _line = re.sub("wire", "logic", _line)
                _line = re.sub("\sreg\s", "logic", _line)
                _line = re.sub(",", "", _line)
                _line = _line + ";"
                instance["io"].append(_line.strip())

    return instance


def get_instance(instance):
    """
    This functions creates the wires declaration and the module instance
    """

    print("INFO: Prepare the testbench")

    mod_inst = ""

    # Print parameter declaration if present
    if instance["parameter"]:
        for param in instance["parameter"]:
            mod_inst += """    """ + param + "\n"
        mod_inst += """\n"""

    # Print input/output declaration if present
    if instance["io"]:
        for io in instance["io"]:
            mod_inst += """    """ + io + "\n"
        mod_inst += """\n"""

    # Write the instance
    mod_inst += """    """ + instance["name"] + " \n"

    # Print parameter instance if present
    if instance["parameter"]:

        mod_inst += """    #(\n"""

        # First get the longest name
        maxlen = 0
        for _, param in enumerate(instance["parameter"]):

            _param = param.split(" ")
            _name = _param[-3]

            if len(_name) > maxlen:
                maxlen = len(_name)

        for ix, param in enumerate(instance["parameter"]):
            # get left and right side around the equal sign
            _param = param.split("=")
            # split over space of the let side ('parameter param_name')
            _param = _param[-2].split(" ")
            # remove empty element in the list
            _param = list(filter(None, _param))
            # grab parameter name, always the last element in the list
            _name = _param[-1]

            _text = "    ." + _name + \
                " " * (maxlen - len(_name)) + \
                " (" + _name + ")"

            mod_inst += _text
            if ix == len(instance["parameter"]) - 1:
                mod_inst += "\n"
            else:
                mod_inst += ",\n"

        mod_inst += "    )\n"

    mod_inst += """    dut \n    (\n"""

    # Print input/output instance if present
    if instance["io"]:

        # First get the longest name
        maxlen = 0
        for _, ios in enumerate(instance["io"]):
            _io = ios.split(" ")
            _name = _io[-1][:-1]
            if len(_name) > maxlen:
                maxlen = len(_name)

        for ix, io in enumerate(instance["io"]):

            _io = io.split(" ")
            # write until the semicolumn
            _name = _io[-1][:-1]

            _text = "    ." + _name + \
                " " * (maxlen - len(_name)) + \
                " (" + _name + ")"

            mod_inst += _text
            if ix == len(instance["io"]) - 1:
                mod_inst += "\n"
            else:
                mod_inst += ",\n"

    mod_inst += """    );\n"""

    return mod_inst

def dump_template(file_name, tmpl):
    """
    Store the template transformated after substitution
    """
    try:
        # Store the testbench
        with open(file_name, "w", encoding="utf-8") as ofile:
            ofile.write(tmpl)
            ofile.close()
    except OSError:
        print("Can't store template")
        sys.exit(1)

    return 0


def print_recommandation(name):
    """
    After the file is written and ready to use, print some
    recommandation to setup and call SVUT flow.
    """

    print("")
    print(f"INFO: Testbench for {name} has been generated in: {name}_testbench.sv")
    print("")
    print("      To launch SVUT, don't forget to setup its environment variable. For instance:")
    print("")
    print("      export SVUT=\"$HOME/.svut\"")
    print("      export PATH=$SVUT:$PATH")
    print("")
    print("      The testbench needs to be tuned to generate clock, reset and test scenarios")
    print("")
    print("      You can setup your fileset in files.f")
    print("")
    print("      Then call SVUT with:")
    print("      svutRun -test my_testbench.sv -define \"DEF1=1;DEF2;DEF3=3\"")
    print("")
    print("      Further information can be found in Github:")
    print("      https://github.com/dpretet/svut#tutorial")
    print("")


def main():
    # Handle the input arguments. A file must be passed
    # and exists in file system
    FILE_NAME = ""

    if len(sys.argv) > 1:
        FILE_NAME = sys.argv[1]
    else:
        print("ERROR: please specify a file to parse")
        sys.exit(1)

    print("INFO: Start to generate the testbench")

    # First extract information from the module to test
    try:
        with open(FILE_NAME, "r", encoding="utf-8") as verilog_module:
            verilog_info = parse_verilog(verilog_module)
    except OSError:
        print(f"ERROR: Can't find file {FILE_NAME} to load...")
        sys.exit(1)

    # Put in shape the module instance and the wire declarations
    module_inst = get_instance(verilog_info)
    # Setup the data to substitute in the template
    tmpl_data = dict(name=verilog_info["name"], module_inst=module_inst)

    # Load the system verilog template and substitute
    tmpl = Path(SCRIPTDIR+"/template.sv", encoding="utf-8").read_text()
    tmpl = Template(tmpl).substitute(tmpl_data)
    dump_template(verilog_info["name"] + "_testbench.sv", tmpl)

    # Load the cpp template and substitute
    tmpl = Path(SCRIPTDIR+"/template.cpp", encoding="utf-8").read_text()
    tmpl = Template(tmpl).substitute(tmpl_data)
    dump_template("sim_main.cpp", tmpl)

    # Print recommandation to users before exiting
    print_recommandation(verilog_info["name"])

    # Create a files.f if doesn't exist, else preserve the existing
    os.system("touch files.f")

    sys.exit(0)

if __name__ == '__main__':
    main()




#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Copyright 2024 The SVUT Authors

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
of the Software, and to permit persons to whom the Software is furnished to do
so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

# pylint: disable=W0621

import os
import sys
import argparse
import filecmp
import subprocess
import shutil
import datetime
from timeit import default_timer as timer
from datetime import timedelta
from pathlib import Path, PosixPath

SCRIPTDIR = os.path.abspath(os.path.dirname(__file__))

def check_arguments(args):
    """
    Verify the arguments are correctly setup
    """

    if "iverilog" in args.simulator or "icarus" in args.simulator:
        print_event("Run with Icarus Verilog")
    elif "verilator" in args.simulator:
        print_event("Run with Verilator")
    else:
        print_event("ERROR: Simulator not supported")
        sys.exit(1)

    if not args.test:
        print_event("ERROR: No testcase or path to testcases passed")
        sys.exit(1)

    for path in args.test:
        error = False
        if not path.exists():
            error = True
            print_event("ERROR: %s does not exist" % path)

    if error:
        sys.exit()

    if args.compile_only and args.run_only:
        print("ERROR: Both compile-only and run-only are used")
        sys.exit(1)

    if (args.compile_only or args.run_only) and args.test=="all":
        print_event("ERROR: compile-only or run-only can't be used with multiple testbenchs")
        sys.exit(1)

    return 0


def check_tb_extension(test : PosixPath):
    """
    Check the extension to be sure it can be run
    """
    if test.suffix not in [".v",".sv"]:
        print("ERROR: Failed to find supported extension. Must use either *.v or *.sv")
        sys.exit(1)


def copy_svut_h():
    """
    First copy svut_h.sv macro in the user folder if not present or different
    """

    # Resolve first the real place h file is located
    # The place is different if the python is directly call or if
    # using the symlink
    org_hfile = SCRIPTDIR + "/svut/svut_h.sv"

    if not os.path.isfile(org_hfile):
        org_hfile =  SCRIPTDIR + "/svut_h.sv"

    curr_hfile = os.getcwd() + "/svut_h.sv"

    if (not os.path.isfile(curr_hfile)) or\
            (not filecmp.cmp(curr_hfile, org_hfile)):
        print("INFO: Copy up-to-date version of svut_h.sv")
        os.system("cp " + org_hfile + " " + os.getcwd())

    return 0


def find_unit_tests(test_dir : PosixPath):
    """
    Parse all unit test files of the current folder
    and return a list of available tests
    """

    supported_prefix = ["tb_", "ts_", "testbench_", "testsuite_", "unit_test_"]
    supported_suffix = ["_unit_test", "_testbench","_testsuite", "_tb", "_ts"]
    files = []

    # Parse the current folder
    for path in Path(test_dir).iterdir():
        # Check only the files
        if not path.is_file():
            continue

        # Files not ending with .sv or .v are skipped
        if path.suffix not in [".sv",".v"]:
            continue

        if path.stem.startswith(tuple(supported_prefix)):
            files.append(path)

        if path.stem.endswith(tuple(supported_suffix)):
            files.append(path)

    # Remove duplicated file if contains both prefix and suffix
    files = list(set(files))

    if not files:
        print("ERROR: Can't find tests to run")
        sys.exit(1)

    return files


def print_banner(tag):
    """
    A banner printed when the flow starts
    """
    print()
    print("""       ______    ____  ________""")
    print("""      / ___/ |  / / / / /_  __/""")
    print("""      \\__ \\| | / / / / / / /  """)
    print("""     ___/ /| |/ / /_/ / / /   """)
    print("""    /____/ |___/\\____/ /_/""")
    print()
    print(f"    {tag}")
    print()

    return 0

def helper(tag):
    """
    Help menu
    """

    print_banner(tag)
    print("    https://github.com/dpretet/svut")
    print()

    return 0


def get_defines(defines):
    """
    Return a string with the list of defines ready to drop in icarus
    """
    simdefs = ""

    if not defines:
        return simdefs

    defs = defines.split(';')

    for _def in defs:
        if _def:
            simdefs += "-D" + _def + " "

    return simdefs


def create_iverilog(args, test):
    """
    Create the Icarus Verilog command to launch the simulation
    """

    cmds = []

    if not os.path.isfile("svut.out"):
        print_event("Testbench executable not found. Will build it")
        args.run_only = False

    # Build testbench executable
    if not args.run_only:

        cmd = "iverilog -g2012 -Wall -o svut.out "

        if args.define:
            cmd += get_defines(args.define)

        if args.dotfile:

            dotfiles = ""

            for dot in args.dotfile:
                if os.path.isfile(dot):
                    dotfiles += dot + " "

            if dotfiles:
                cmd += "-f " + dotfiles + " "

        if args.include:
            incs = " ".join(args.include)
            cmd += "-I " + incs + " "

        cmd += test + " "
        cmds.append(cmd)

    # Execute testbench
    if not args.compile_only:

        cmd = "vvp "
        if args.vpi:
            cmd += args.vpi + " "

        cmd += "svut.out "
        cmds.append(cmd)

    return cmds


def create_verilator(args, test):
    """
    Create the Verilator command to launch the simulation
    """

    testname = os.path.basename(test).split(".")[0]

    cmds = []

    if not os.path.isfile("build/V" + testname + ".mk"):
        print_event("Testbench executable not found. Will build it")
        args.run_only = False


    # Build testbench executable
    if not args.run_only:

        cmd = """verilator -Wall --trace --Mdir build +1800-2012ext+sv """
        cmd += """+1800-2005ext+v -Wno-STMTDLY -Wno-UNUSED -Wno-UNDRIVEN -Wno-PINCONNECTEMPTY """
        cmd += """-Wpedantic -Wno-VARHIDDEN -Wno-lint """

        if args.define:
            cmd += get_defines(args.define)

        if args.dotfile:

            dotfiles = ""

            for dot in args.dotfile:
                if os.path.isfile(dot):
                    dotfiles += dot + " "

            if dotfiles:
                cmd += "-f " + dotfiles + " "

        if args.include:
            for inc in args.include:
                cmd += "+incdir+" + inc + " "

        cmd += "-cc --exe --build -j --top-module " + testname + " "
        cmd += test + " " + args.main
        cmds.append(cmd)

    # Execution command
    if not args.compile_only:
        cmd = "build/V" + testname
        cmds.append(cmd)

    return cmds


def print_event(event):
    """
    Print an event during SVUT execution
    TODO: manage severity/verbosity level
    """

    time = datetime.datetime.now().time().strftime('%H:%M:%S')

    print("SVUT (@ " + time + ") " + event, flush=True)
    print("")

    return 0


def get_git_tag():
    """
    Return current SVUT version
    """

    # Handles an environment without Git installed
    git_path = shutil.which("git")
    if git_path is None:
        return ""

    curr_path = os.getcwd()
    os.chdir(SCRIPTDIR)

    try:
        git_tag = subprocess.check_output(["git", "describe", "--tags", "--abbrev=0"])
        git_tag = git_tag.strip().decode('ascii')
    except subprocess.CalledProcessError as err:
        print("WARNING: Can't get last git tag. Will return v0.0.0")
        git_tag = "v0.0.0"

    os.chdir(curr_path)
    return git_tag

def get_test_dir(input: list) -> PosixPath:
    if len(input) == 1 and input[0].is_dir():
        return input[0]

    return None


def main():
    """
    Main function
    """

    parser = argparse.ArgumentParser(description='SystemVerilog Unit Test Flow')

    # SVUT options

    parser.add_argument('-sim', dest='simulator', type=str, default="icarus",
                        help='The simulator to use, icarus or verilator.')

    parser.add_argument('-test', dest='test', type=str, default=[os.getcwd()], nargs="*",
                        help='Unit test to run. A file,list of files or path to test files')

    parser.add_argument('-no-splash', dest='splash', default=False, action='store_true',
                        help='Don\'t print the banner when executing')

    parser.add_argument('-version', dest='version', action='store_true',
                        default="", help='Print version menu')

    # Simulator options

    parser.add_argument('-f', dest='dotfile', type=str, default=["files.f"], nargs="*",
                        help="A dot file (*.f) with incdir, define and file path")

    parser.add_argument('-include', dest='include', type=str, nargs="*",
                        default="", help='Specify an include folder; can be used along a dotfile')

    parser.add_argument('-main', dest='main', type=str, default="sim_main.cpp",
                        help='Verilator main cpp file, like sim_main.cpp')

    parser.add_argument('-define', dest='define', type=str, default="",
                        help='''A list of define separated by ; \
                            ex: -define "DEF1=2;DEF2;DEF3=3"''')

    parser.add_argument('-vpi', dest='vpi', type=str, default="",
                        help='''A string of arguments passed as is to Icarus (only), separated \
                                by a space ex: -vpi "-M. -mMyVPI"''')

    # SVUT Execution options

    parser.add_argument('-run-only', dest='run_only', default=False, action='store_true',
                        help='Only run existing executable but build it if not present')

    parser.add_argument('-compile-only', dest='compile_only', default=False, action='store_true',
                        help='Only prepare the testbench executable')

    parser.add_argument('-dry-run', dest='dry', default=False, action='store_true',
                        help='Just print the command, don\'t execute')


    args = parser.parse_args()
    args.test = [Path(x) for x in args.test]

    git_tag = get_git_tag()

    if args.version:
        helper(git_tag)
        sys.exit(0)

    if not args.splash:
        print_banner(git_tag)

    # Lower the simulator name to ease checking
    args.simulator = args.simulator.lower()

    # Check arguments consistency
    check_arguments(args)

    # If the user specifies a directory to look for files, test_dir will point to this directory
    # If the user does not specify any test files or directy at all, test_dir will point to the CWD
    # If the user specifies one or more test files, test_dir is none and no searching will take place
    test_dir = get_test_dir(args.test)
    if test_dir:
        args.test = find_unit_tests(test_dir)

    # Copy svut_h.sv if not present or not up-to-date
    copy_svut_h()

    cmdret = 0

    start = timer()

    test : PosixPath
    for test in args.test:

        check_tb_extension(test)

        if "iverilog" in args.simulator or "icarus" in args.simulator:
            cmds = create_iverilog(args, str(test))

        elif "verilator" in args.simulator:
            cmds = create_verilator(args, str(test))

        print_event("Start " + test.name)

        # Execute commands one by one
        for cmd in cmds:

            print_event(cmd)

            if not args.dry:
                if os.system(cmd):
                    cmdret += 1
                    print("ERROR: Command failed: " + cmd)
                    break

        print_event("Stop " + test.name)

    end = timer()
    print_event("Elapsed time: " + str(timedelta(seconds=end-start)))
    print()

    sys.exit(cmdret)

if __name__ == '__main__':
    main()



"""Setup module for SVUT.
"""

from setuptools import setup, find_namespace_packages
import pathlib

here = pathlib.Path(__file__).parent.resolve()

long_description = (here / "README.md").read_text(encoding="utf-8")


setup(
    name="svut",
    version="1.9.3",
    description="SystemVerilog Unit Test (SVUT)",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/dpretet/svut",
    author="Damien Pretet",
    author_email="damien.pretet@me.com",
    classifiers=[
        "Development Status :: 3 - Production/Stable",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Code Generators",
        "Topic :: System :: Hardware",
        "Topic :: Software Development :: Testing",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
    ],
    keywords="Verilog Testbench Digital-Design",
    packages=find_namespace_packages(),
    include_package_data=True,
    python_requires=">=3.7, <4",
    entry_points={
        "console_scripts": [
            "svutRun=svut.svutRun:main",
            "svutCreate=svut.svutCreate:main",
        ],
    },
    project_urls={
        "Bug Reports": "https://github.com/dpretet/svut/issues",
        "Say Thanks!": "http://saythanks.io/to/example",
        "Source": "https://github.com/dpretet/svut/",
    },
)
